/*	marqtt.go - M.A.R-T>Y. MQTT bridge daemon written in Go
	Copyright (C) 2023-2024 hoff.industries

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>. */

package main

import (
	"os"
	"fmt"
	"hoff.industries/marty/libmmp"
	"hoff.industries/marty/libmarty"
	"hoff.industries/marty/mendpoints"
	mqtt "github.com/eclipse/paho.mqtt.golang"
)

var TARGET_PROTOCOL = []int{2, 0, 0}
var INSTANCE_VERSION = []int{3, 0, 0}

const INSTANCE_NAME string = "m-arqtt"
const INSTANCE_TITLE string = "MQTT Bridge Service"

const TABLE_SIZE int = 20

var Endpoints = []libmarty.MartyClientEndpoint{

	{
		Endpoint: mendpoints.MARQTT_EPT_SUBSCRIBE,
		Name: "subscribe",
		Min: 6,
		Max: 6,
		IsAscii: true,
	},

	{
		Endpoint: mendpoints.MARQTT_EPT_UNSUBSCRIBE,
		Name: "unsubscribe",
		Min: 6,
		Max: 6,
		IsAscii: true,
	},

	{
		Endpoint: mendpoints.MARQTT_EPT_PUBLISH,
		Name: "publish",
		Min: 7,
		Max: 7,
		IsAscii: true,
	},
}

var MaxEndpoint = len(Endpoints)

var disconnectHandler mqtt.ConnectionLostHandler = func(client mqtt.Client, err error) {

	var reconnectStatus bool = false
	var reconnectToken mqtt.Token;

	fmt.Printf("MQTT Connection lost: %v", err)

 	for !reconnectStatus {
		reconnectToken = client.Connect()
		reconnectToken.Wait()

		reconnectStatus = reconnectToken.Error() == nil
	}

	fmt.Printf("MQTT Reconnection success!\n")
}

func main() {
	var publishToken mqtt.Token

	var data libmarty.MartyClientData
	var status libmmp.MMStatus

	var txPacket libmmp.MMPacket
	var rxPacket *libmmp.MMPacket = nil

	var applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["SUCCESS"]

	var hostname string = ""
	var callError error = nil

	var endpointConversion int = -1
	var endpoint *libmarty.MartyClientEndpoint = nil

	var mqttClient mqtt.Client
	var mqttOpts *mqtt.ClientOptions
	var mqttToken mqtt.Token

	//var subscribeTable map[string][]*libmarty.MartyClientParameters
	//var deliveryHandle *libmarty.MartyMessageHandle = nil

	libmarty.Marty_client_data_factory(&data)

	libmmp.Mmstatus_factory(&status)
	libmmp.Mmpacket_factory(&txPacket, true)

	data.VersionString = fmt.Sprintf("%s v%d.%d.%d", INSTANCE_NAME,
		INSTANCE_VERSION[0], INSTANCE_VERSION[1], INSTANCE_VERSION[2])

	fmt.Printf("%s[%s]%s Starting %s %s - %s\n",
		libmarty.LoggerColours[libmarty.LOG_INFO], libmarty.MartyFlavour,
			libmarty.ClearColour, libmarty.LIBMARTY_AUTHOR, INSTANCE_TITLE, data.VersionString)

	hostname, callError = os.Hostname()
	if (callError != nil) {

		os.Exit(libmarty.EXIT_ERR_HOSTNAME)
	}

	data.ConnType = libmarty.CONN_UNIX
	data.InitParameters.Auth = ""
	data.InitParameters.Priority = 0
	data.InitParameters.Host = hostname
	data.InitParameters.Name = "marqtt"

	mqttOpts = mqtt.NewClientOptions()

	mqttOpts.AddBroker(fmt.Sprintf("tcp://%s:%d",
		"127.0.0.1", 9961))

	mqttOpts.SetClientID("marqtt")
	mqttOpts.SetUsername("")
	mqttOpts.SetPassword("")
	mqttOpts.OnConnectionLost = disconnectHandler
	//mqttOpts.SetDefaultPublishHandler(messageHandler)

	mqttClient = mqtt.NewClient(mqttOpts)

	mqttToken = mqttClient.Connect()
	mqttToken.Wait()

	if mqttToken.Error() != nil {

		libmarty.Marty_full_log("MQTT initialisation failure!\n", libmarty.LOG_CRITICAL, &data)
		os.Exit(libmarty.EXIT_ERR_INIT)
	}

	libmarty.Marty_full_log("MQTT initialisation success!\n", libmarty.LOG_STARTUP, &data)

	if (libmarty.Spawn_marty_socket(&data) != libmmp.NONE) {

		libmarty.Marty_full_log("Socket spawn failure!\n",
			libmarty.LOG_CRITICAL, &data)
		os.Exit(libmarty.EXIT_ERR_SPAWN)
	}

	if (libmarty.Marty_init(&data) != libmmp.NONE) {

		libmarty.Marty_full_log("MMP initialisation failure!\n",
			libmarty.LOG_CRITICAL, &data)
		os.Exit(libmarty.EXIT_ERR_INIT)
	};

	libmarty.Marty_full_log("MMP initialisation success!\n", libmarty.LOG_STARTUP, &data)

	go libmarty.Marty_network_loop(&data)
	libmarty.Marty_full_log("Network loop started.\n", libmarty.LOG_STARTUP, &data)

	for true {

		libmmp.Mmpacket_reset(&txPacket, true)
		applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["SUCCESS"]

		if (rxPacket != nil) {
			libmmp.Mmpacket_reset(rxPacket, true)
		}

		rxPacket = nil

		select {
			//case deliveryHandle = <- data.DeliveryChannel:
			case _ = <- data.DeliveryChannel:
				continue
			// listen for new SEND messages from client loop on channel
			case rxPacket = <- data.MessageChannel:
		}

		if (rxPacket == nil) {
			libmarty.Marty_full_log("Received NULL pointer from ITC channel!\n", libmarty.LOG_WARNING, &data)
			continue
		}

		endpointConversion = int(rxPacket.Payloads[libmarty.PAYL_IDX_ENDPOINT][0])

		if (endpointConversion > MaxEndpoint || endpointConversion < 0) {

			libmarty.Marty_full_log("Requested endpoint outside my range!\n", libmarty.LOG_WARNING, &data)
			applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["ENDPOINT"]

			goto JMP_TX_STATUS
		}

		endpoint = &Endpoints[endpointConversion]

		if (len(rxPacket.Payloads) < endpoint.Min) {
			libmarty.Marty_full_log("Request payload count less than expected!\n", libmarty.LOG_WARNING, &data)

			applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["TOO_FEW"]

			goto JMP_TX_STATUS
		}

		if (len(rxPacket.Payloads) > endpoint.Max) {
			libmarty.Marty_full_log("Request payload count more than expected!\n", libmarty.LOG_WARNING, &data)

			applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["TOO_MANY"]

			goto JMP_TX_STATUS
		}

		switch (endpointConversion) {
			case mendpoints.MARQTT_EPT_SUBSCRIBE:
				break

			case mendpoints.MARQTT_EPT_UNSUBSCRIBE:
				break

			case mendpoints.MARQTT_EPT_PUBLISH:

				publishToken = mqttClient.Publish(libmmp.Payload_2_str(rxPacket.Payloads[5]),
					0, false, libmmp.Payload_2_str(rxPacket.Payloads[6]))

				publishToken.Wait()

				libmarty.Marty_full_log("Published a message to the topic ", libmarty.LOG_INFO, &data)
				fmt.Printf("'%s'\n", libmmp.Payload_2_str(rxPacket.Payloads[5]))
		}

		JMP_TX_STATUS:

		libmarty.Marty_set_application(&txPacket, &status, rxPacket.Payloads[0], applicationStatus)
		if (status.Error != libmmp.NONE) {
			libmarty.Marty_full_log("Couldn't set autopacket!\n", libmarty.LOG_ERROR, &data)
			continue
		}

		libmmp.Serialize(&txPacket, &status)
		if (status.Error != libmmp.NONE) {
			libmarty.Marty_full_log("Couldn't serialize autopacket!\n", libmarty.LOG_ERROR, &data)
			fmt.Printf("%d\n", status.Error)
			continue
		}

		libmarty.Marty_net_write(&txPacket, &status, &data)
	}
}
